'''
Module that contains the various data types:

1. Target -- Mutable data store for a single Target value i.e. one training \
example.
2. TargetCollection -- Mutable data store for Target data types.  i.e. A \
data store that contains multiple Target instances.
'''

from collections.abc import MutableMapping
from collections import OrderedDict, defaultdict
import copy
import json
from pathlib import Path
import random as rand
from typing import List, Callable, Union, Dict, Tuple, Any, Optional, Set
import warnings

import numpy as np
import pandas as pd
from sklearn import metrics
from sklearn.model_selection import train_test_split
import seaborn as sns

from bella.tokenisers import whitespace
from bella.stanford_tools import constituency_parse

BELLA_DATASET_DIR: Path = Path.home().joinpath('.Bella', 'Datasets')

class Target(MutableMapping):
    '''
    Mutable data store for a single Target value. This should be used as the
    value for Target information where it contains all data required to be
    classified as Target data for Target based sentiment classification.

    Overrides collections.abs.MutableMapping abstract class.

    Reference on how I created the class:
    http://www.kr41.net/2016/03-23-dont_inherit_python_builtin_dict_type.html
    https://docs.python.org/3/library/collections.abc.html

    Functions changed compared to normal:

    1. __delitem__ -- Will only delete the `target_id` key.
    2. __eq__ -- Two Targets are the same if they either have the same `id` or \
    they have the same values to the minimum keys \
    ['spans', 'text', 'target', 'sentiment']
    3. __setitem__ -- Only allows you to add/modify the `predicted` key which \
    represents the predicted sentiment for the Target instance.
    '''

    def __init__(self, spans, target_id, target, text, sentiment, predicted=None,
                 sentence_id=None, category=None, augmented=None, 
                 transfer_data=None, original_target_id=None):
        '''
        :param target: Target that the sentiment is about. e.g. Iphone
        :param sentiment: sentiment of the target.
        :param text: The text context that the target and sentiment is within.
        :param target_id: Unique ID. Has to be Unique within the 
                          TargetCollection if it is put into a TargetCollection.
        :param spans: List of tuples where each tuple is of length 2 where they 
                      contain the exclusive range of an instance of Target word 
                      in the Text context. The reason it is a list if because 
                      the Target word can be mentioned more than once e.g. 
                      `The Iphone was great but the iphone is small`. The first 
                      Int in the tuple has to be less than the second Int.
        :param predicted: If given adds the predicted sentiment value.
        :param sentence_id: Unique ID of the sentence that the target is 
                            within. More than one target can have the same 
                            sentence.
        :param category: In some datasets there is category information where 
                         the target is assigned a category. This comes from 
                         the SemEval 2015 restaurant dataset.
        :param augmented: Whether or not the data comes from an augmented 
                          dataset or has been produced from an augmented 
                          function.
        :param transfer_data: Whether or not the data comes from a transfer 
                              dataset.
        :param original_target_id: If the augmented field is True then the 
                                   original_target_id has to exist, as this 
                                   will inform what the original target 
                                   was used to create this augmented target 
                                   data. 
        :type target: String
        :type sentiment: String or Int (Based on annotation schema)
        :type text: String
        :type target_id: String
        :type spans: list
        :type predicted: Same type as sentiment. Default None (Optional)
        :type sentence_id: String. Default None (Optional)
        :type category: String. Default None (Optional)
        :type augmented: bool. Default None (Optional)
        :type transfer_data: bool. Default None (Optional)
        :type original_target_id: String
        :returns: Nothing. Constructor.
        :rtype: None
        '''
        if not isinstance(target_id, str):
            raise TypeError('The target ID has to be of type String and not {}'\
                            .format(type(target_id)))
        if not isinstance(target, str):
            raise TypeError('The target has to be of type String and not {}'\
                            .format(type(target)))
        if not isinstance(text, str):
            raise TypeError('The text has to be of type String and not {}'\
                            .format(type(text)))
        if not isinstance(sentiment, (str, int)):
            raise TypeError('The sentiment has to be of type String or Int and '\
                            'not {}'.format(type(sentiment)))
        if not isinstance(spans, list):
            raise TypeError('The spans has to be of type list and not {}'\
                            .format(type(spans)))
        else:
            if len(spans) < 1:
                raise TypeError('spans has to contain at least one tuple not '\
                                'None')
            else:
                for span in spans:
                    if not isinstance(span, tuple):
                        raise TypeError('Spans has to be a list of tuples not {}'\
                                        .format(type(span)))
                    if len(span) != 2:
                        raise ValueError('Spans must contain tuples of length'\
                                         ' 2 not {}'.format(spans))
                    if not isinstance(span[0], int) or \
                       not isinstance(span[1], int):
                        raise TypeError('spans must be made of tuple containing '\
                                        'two Integers not {}'.format(span))
                    if span[1] <= span[0]:
                        raise ValueError('The first integer in a span must be '\
                                         'less than the second integer {}'\
                                         .format(span))
        temp_dict = dict(spans=spans, target_id=target_id, target=target,
                         text=text, sentiment=sentiment)
        if sentence_id is not None:
            if not isinstance(sentence_id, str):
                raise TypeError('`sentence_id` has to be a String and not {}'\
                                .format(type(sentence_id)))
            temp_dict['sentence_id'] = sentence_id

        if category is not None:
            temp_dict['category'] = category
        if augmented is not None:
            temp_dict['augmented'] = augmented
            if original_target_id is None:
                aug_data_error = ('Cannot create a Target that is augmented '
                                  'without having the original target_id of the'
                                  ' target data that was augmented to create '
                                  'this target data')
                raise ValueError(aug_data_error)
            
        if original_target_id is not None:
            temp_dict['original_target_id'] = original_target_id
        if transfer_data is not None:
            temp_dict['transfer_data'] = transfer_data

        self._storage = temp_dict
        if predicted is not None:
            self['predicted'] = predicted


    def __getitem__(self, key):
        return self._storage[key]

    def __iter__(self):
        return iter(self._storage)

    def __len__(self):
        return len(self._storage)

    def __delitem__(self, key):
        '''
        To ensure that the Target class maintains the minimum Keys and Values
        to allow an instance to be used in Target based machine learning. The
        key and associated values that can be deleted are limited to:

        1. target_id

        :param key: The key and associated value to delete from the store.
        :returns: Updates the data store by removing key and value.
        :rtype: None
        '''

        accepted_keys = set(['target_id'])
        if key not in accepted_keys:
            raise KeyError('The only keys that can be deleted are the '\
                           'following: {} the key you wish to delete {}'\
                           .format(accepted_keys, key))
        del self._storage[key]

    def __setitem__(self, key, value):
        '''
        :param key: key (Only store values for `predicted` key)
        :param value: Predicted sentiment value which has to be the same data \
        type as the `sentiment` value.
        :type key: String (`predicted` is the only key accepted at the moment)
        :type value: Int or String.
        :returns: Nothing. Adds the predicted sentiment of the Target.
        :rtype: None.
        '''

        if key != 'predicted':
            raise KeyError('The Only key that can be changed is the `predicted`'\
                           ' key not {}'.format(key))
        #raise_type = False
        #sent_value = self._storage['sentiment']
        #if isinstance(sent_value, int):
        #    if not isinstance(value, (int, np.int32, np.int64)):
        #        raise_type = True
        #elif not isinstance(value, type(sent_value)):
        #    raise_type = True

        #if raise_type:
        #    raise TypeError('Value to be stored for the `predicted` sentiment '\
        #                    'has to be the same data type as the sentiment '\
        #                    'value {} and not {}.'\
        #                    .format(sent_value, type(value)))
        self._storage[key] = value

    def __repr__(self):
        '''
        :returns: String returned is what user see when the instance is \
        printed or printed within a interpreter.
        :rtype: String
        '''

        return 'Target({})'.format(self._storage)

    def __eq__(self, other):
        '''
        Two Target instances are equal if they are both Target instances and
        one of the following conditions

        1. They have the same target_id (This is preferred)
        2. The minimum keys that all targets have to have \
        ['spans', 'text', 'target', 'sentiment'] are all equal.

        :param other: The target instance that is being compare to the current \
        target instance.
        :type other: Target
        :returns: True if they are equal else False.
        :rtype: bool
        '''

        if not isinstance(other, Target):
            return False
        if 'target_id' in self and 'target_id' in other:
            if self['target_id'] != other['target_id']:
                return False
        else:
            minimum_keys = ['spans', 'text', 'target', 'sentiment']
            for key in minimum_keys:
                if not self[key] == other[key]:
                    return False
        return True
    def __array__(self):
        '''
        Function for converting it to a numpy array
        '''
        return np.asarray(dict(self))

class TargetCollection(MutableMapping):
    '''
    Mutable data store for Target data types.  i.e. A data store that contains
    multiple Target instances. This collection ensures that there are no two
    Target instances stored that have the same ID this is because the storage
    of the collection is an OrderedDict.

    Overrides collections.abs.MutableMapping abstract class.

    Functions:

    1. add -- Given a Target instance with an `id` key adds it to the data 
       store.
    2. data -- Returns all of the Target instances stored as a list of Target 
       instances.
    3. stored_sentiments -- Returns a set of unique sentiments stored.
    4. sentiment_data -- Returns the list of all sentiment values stored in 
       the Target instances stored.
    5. add_pred_sentiment -- Adds a list of predicted sentiment values to the 
       Target instances stored.
    6. confusion_matrix -- Returns the confusion matrix between the True and 
       predicted sentiment values.
    7. subset_by_sentiment -- Creates a new TargetCollection based on the 
       number of unique sentiments in a sentence.
    8. to_json_file -- Returns a Path to a json file that has stored the 
       data as a sample json encoded per line. If the split argument is set 
       will return two Paths the first being the training file and the second 
       the test.
    9. categories_targets -- Returns two dictionaries. The first is a category
       to target list and the second is a target to category list. Their is 
       the option to have a coarse grained cateogires option.
    10. target_set -- Returns a set of all the targets within this dataset.
    11. subset_by_ids -- Returns a TargetCollection which is a subset of the 
        current one. Where the samples included in the subset are those with 
        a `target_id` that is within the ids List.
    12. dataset_metric_scores -- Given a metric like accuracy returns all of 
        the metric scores for the dataset. This assumes that the dataset has 
        the `predicted` sentiment slot assigned.
    13. subset_by_targets -- Will subset the data so that only data points that 
        contain the given targets are within the TargetCollection.

    Static functions:
    
    1. target_targets -- Returns a dictionary of targets as keys and values 
       are a list of realted targets.
    2. split_dataset -- Returns the given dataset into two, the train and test 
                        based on the test_split fraction given.

    Attributes:
    1. grouped_sentences -- A dictionary of sentence_id as keys and a list of 
       target instances that have the same sentence_id as values.
    2. grouped_sentiments -- A dictionary where the keys are all possible 
       sentiment values, the value is a list of Target(s) that only have that 
       associated sentiment value.
    3. grouped_distinct_sentiments -- A dictionary where the keys are the 
       number of distinct sentiments that the targets have in the associated 
       value. A target has two distinct sentiments if the associated sentence 
       has contains targets that take on one of two sentiment where at least 
       one target has a different sentiment to the rest. The targets in the 
       values of this dictionary are stored as a list of Target
    '''

    def __init__(self, list_of_target=None, name: Optional[str]=None):
        '''
        :param list_of_target: An interator of Target instances e.g. a List of 
                               Target instances.
        :type list_of_target: Iterable. Default None (Optional)
        :returns: Nothing. Constructor.
        :rtype: None
        '''

        self._storage = OrderedDict()
        if list_of_target is not None:
            if not hasattr(list_of_target, '__iter__'):
                raise TypeError('The list_of_target argument has to be iterable')
            for target in list_of_target:
                self.add(target)
        self.name = name if name is not None else 'TargetCollection'
        self._grouped_sentences = None
        self._grouped_sentiments = None
        self._grouped_distinct_sentiments = None

    def __getitem__(self, key):
        return self._storage[key]

    def __setitem__(self, key, value):
        '''
        If key already exists will raise KeyError.

        :param key: key that stores the index to the value
        :param value: value to store at the keys location
        :type key: hashable object
        :type value: Target
        :returns: Nothing. Adds data to the collection
        :rtype: None.
        '''

        # Required to make sure the grouped_sentences, grouped_sentiments,
        # and grouped_distinct_sentiments get recomputed instead of cached.
        self._data_has_changed = True
        self._data_has_changed_gs = True
        self._data_has_changed_gds = True
        if not isinstance(value, Target):
            raise TypeError('All values in this store have to be of type '\
                            'Target not {}'.format(type(value)))
        if key in self._storage:
            raise KeyError('This key: `{}` already exists with value `{}` '\
                           'value that for the same key is `{}`'\
                           .format(key, self._storage[key], value))
        temp_value = copy.deepcopy(value)
        # As the id will be saved as the key no longer needed in the target
        # instance (value). However if the key does not match the `target_id`
        # raise KeyError
        if 'target_id' in value:
            if value['target_id'] != key:
                raise KeyError('Cannot add this to the data store as the key {}'\
                               ' is not the same as the `target_id` in the Target'\
                               ' instance value {}'.format(key, value))
            del temp_value['target_id']
        self._storage[key] = temp_value

    def __delitem__(self, key):
        del self._storage[key]

    def __iter__(self):
        return iter(self._storage)

    def __len__(self):
        return len(self._storage)

    def add(self, value):
        '''
        Adds the Target instance to the data store without having to extract
        out the target_id of the target if using __setitem__

        :Example:

        >>> target = Target([(10, 16)], '1', 'Iphone',
                            'text with Iphone', 0)
        >>> target_col = TargetCollection()
        # Add method is simpler to use than __setitem__
        >>> target_col.add(target)
        # Example of the __setitem__ method
        >>> target_col[target['target_id']] = target

        :param value: Target instance with a `target_id` key
        :type value: Target
        :returns: Nothing. Adds the target instance to the data store.
        :rtype: None
        '''

        if not isinstance(value, Target):
            raise TypeError('All values in this store have to be of type '\
                            'Target not {}'.format(type(value)))
        if 'target_id' not in value:
            raise ValueError('The Target instance given {} does not have a '\
                             'target_id'.format(value))
        self[value['target_id']] = value

    def data(self):
        '''
        :returns: a list of all the Target instances stored.
        :rtype: list
        '''

        _data = []
        for _id, target_data in self.items():
            data_dict = {**target_data}
            data_dict['target_id'] = _id
            _data.append(Target(**data_dict))
        return _data

    def data_dict(self):
        '''
        :returns: Same as the data function but returns dicts instead of \
        Targets
        :rtype: list
        '''

        _data = []
        for _id, target_data in self.items():
            data_dict = {**target_data}
            data_dict['target_id'] = _id
            _data.append(data_dict)
        return _data

    def stored_sentiments(self) -> Set[Any]:
        '''
        :returns: A set of all unique sentiment values of the target instances \
        in the data store.
        :rtype: set
        '''

        unique_sentiments = set()
        for target_data in self.values():
            unique_sentiments.add(target_data['sentiment'])
        return unique_sentiments

    def sentiment_data(self, mapper=None, sentiment_field='sentiment'):
        '''
        :param mapper: A dictionary that maps the keys to the values where the \
        keys are the current unique sentiment values of the target instances \
        stored
        :param sentiment_field: Determines if it should return True sentiment \
        of the Targets `sentiment` or to return the predicted value `predicted`
        :type mapper: dict
        :type sentiment_field: String. Default `sentiment` (True values)
        :returns: a list of the sentiment value for each Target instance stored.
        :rtype: list

        :Example of using the mapper:
        >>> target_col = TargetCollection([Target([(10, 16)], '1', 'Iphone',
                                                  'text with Iphone', 'pos')])
        >>> target_col.add(Target([(10, 15)], '2', 'Pixel',
                                  'text with Pixel', 'neg'))
        # Get the unique sentiment values for each target instance
        >>> map_keys = target_col.stored_sentiments()
        >>> print(map_keys)
        >>> ['pos', 'neg']
        >>> mapper = {'pos' : 1, 'neg' : -1}
        >>> print(target_col.sentiment_data(mapper=mapper))
        >>> 1, -1
        '''

        allowed_fields = set(['sentiment', 'predicted'])
        if sentiment_field not in allowed_fields:
            raise ValueError('The `sentiment_field` has to be one of the '\
                             'following values {} and not {}'\
                             .format(allowed_fields, sentiment_field))

        if mapper is not None:
            if not isinstance(mapper, dict):
                raise TypeError('The mapper has to be of type dict and not {}'\
                                .format(type(mapper)))
            allowed_keys = self.stored_sentiments()
            if len(mapper) != len(allowed_keys):
                raise ValueError('The mapper has to contain a mapping for each '\
                                 'unique sentiment value {} and not a subset '\
                                 'given {}'.format(allowed_keys, mapper.keys()))
            for map_key in mapper:
                if map_key not in allowed_keys:
                    raise ValueError('The mappings are not correct. The map '\
                                     'key {} does not exist in the unique '\
                                     'sentiment values in the store {}'\
                                     .format(map_key, allowed_keys))
            return [mapper[target_data[sentiment_field]]\
                    for target_data in self.values()]

        return [target_data[sentiment_field] for target_data in self.values()]

    def add_id_pred(self, id_pred):
        count = 0
        for targ_id in self:
            if targ_id in id_pred:
                self[targ_id]['predicted'] = id_pred[targ_id]
                count += 1
        if count != len(self):
            raise ValueError('We have only added {} predictions to {} targets'\
                             .format(count, len(self)))

    def add_pred_sentiment(self, sent_preds: Union[List[Any], np.ndarray], 
                           mapper: Optional[Dict[Any, Any]] = None) -> None:
        '''
        :param sent_preds: A list of predicted sentiments for all Target 
                           instances stored or a numpy array where columns are 
                           number of different predicted runs and the rows 
                           represent the associated Target instance.
        :param mapper: A dictionary mapping the predicted sentiment to 
                       alternative values e.g. Integer values to String values.
        :type sent_preds: list or numpy array
        :type mapper: dict
        :returns: Nothing. Adds the predicted sentiments to the Target 
                   instances stored.
        '''

        if len(sent_preds) != len(self):
            raise ValueError('The length of the predicted sentiments {} is not '\
                             'equal to the number Target instances stored {}'\
                             .format(len(sent_preds), len(self)))
        for index, target in enumerate(self.data()):
            predicted_sent = sent_preds[index]
            if mapper is not None:
                predicted_sent = mapper[predicted_sent]
            target_id = target['target_id']
            self._storage[target_id]['predicted'] = predicted_sent

    def confusion_matrix(self, plot=False, norm=False):
        '''
        :param plot: To return a heatmap of the confusion matrix.
        :param norm: Normalise the values in the confusion matrix
        :type plot: bool. Default False
        :type norm: bool. Default False
        :returns: A tuple of length two. 1. the confusion matrix \
        2. The plot of the confusion matrix if plot is True else \
        None.
        :rtype: tuple
        '''

        sentiment_values = sorted(self.stored_sentiments())
        true_values = self.sentiment_data()
        pred_values = self.sentiment_data(sentiment_field='predicted')
        conf_matrix = metrics.confusion_matrix(true_values, pred_values,
                                               labels=sentiment_values)
        if norm:
            conf_matrix = conf_matrix / conf_matrix.sum()
        conf_matrix = pd.DataFrame(conf_matrix, columns=sentiment_values,
                                   index=sentiment_values)
        ax = None
        if plot:
            if norm:
                ax = sns.heatmap(conf_matrix, annot=True, fmt='.2f')
            else:
                ax = sns.heatmap(conf_matrix, annot=True, fmt='d')
        return conf_matrix, ax

    def subset_by_sentiment(self, num_unique_sentiments):
        '''
        Creates a subset based on the number of unique sentiment values per
        sentence. E.g. if num_unique_sentiments = 2 then it will
        return all the Target instances where each target intance has at least
        two target instances per sentence and those targets can have only one
        of two sentiment values. This can be used to test how well a method
        can extract exact sentiment information for the associated target.

        NOTE: Requires that all Target instances stored contain a sentence_id.

        :param num_unique_sentiments: Integer specifying the number of unique \
        sentiments in the target instances per sentence.
        :type num_unique_sentiments: int
        :returns: A subset based on the number of unique sentiments per sentence.
        :rtype: TargetCollection
        '''

        all_relevent_targets = []
        for targets in self.grouped_sentences.values():
            target_col = TargetCollection(targets)
            if len(target_col.stored_sentiments()) == num_unique_sentiments:
                all_relevent_targets.extend(targets)
        return TargetCollection(all_relevent_targets)

    def subset_by_targets(self, target_set: Set[str]):
        '''
        Will subset the data so that only data points that contain the given 
        targets are within the TargetCollection.
        
        :param target_set: A list of targets to be used to sub-sample 
                           the data collection from.
        :returns: The current target collection with data points that only 
                  have targets that are from the target set.
        :rtype: TargetCollection
        '''

        target_data = []
        for data in self.data_dict():
            if data['target'].lower() in target_set:
                target_data.append(Target(**data))
        return TargetCollection(target_data)

    def subset_by_sentence_length(self, length_condition):

        all_relevent_targets = []
        for target in self.data():
            target_text = target['text']
            if length_condition(target_text):
                all_relevent_targets.append(target)
        return TargetCollection(all_relevent_targets)

    @staticmethod
    def split_dataset(data: 'TargetCollection', test_split: float, 
                      random: bool = False
                      ) -> Tuple['TargetCollection', 'TargetCollection']:
        '''
        It will split the data into training and test splits, where the amount of 
        data in the test is defined by the test_split fraction (amount of training 
        data will be 1 - test_split).

        :param data: data to split
        :param test_split: The amount of data to give to the test split.
        :param random: If the splitting should be random rather than the last 
                    test split of the data being test and the first 
                    (1 - test split) being the training data.
        :returns: The training data and test data as a tuple.
        '''
        target_data = data.data_dict()
        data_size = len(target_data)
        test_data_size = int(data_size * test_split)
        test_data = []
        train_data = []
        if random:
            start_index = rand.randint(0, data_size)
            end_index = start_index + test_data_size
            if end_index > data_size:
                index_diff = end_index - data_size
                start_index -= index_diff
                end_index = data_size
            test_data = target_data[start_index: end_index]
            train_data = target_data[end_index:]
            train_data.extend(target_data[: start_index])
        else:
            test_data = target_data[: test_data_size]
            train_data = target_data[test_data_size:]
        train_data = [Target(**t_data) for t_data in train_data]
        test_data = [Target(**t_data) for t_data in test_data]
        return (TargetCollection(train_data), TargetCollection(test_data))

    def to_json_file(self, dataset_name: Union[str, List[str]], 
                     split: Optional[float] = None, cache: bool = True, 
                     group_by_sentence: bool = False, random: bool = False
                     ) -> Union[Path, List[Path]]:
        '''
        Returns a Path to a json file that has stored the data as a sample json 
        encoded per line. If the split argument is set will return two Paths 
        the first being the training file and the second the test.

        The Path does not need to be specified as it saves it to the 
        `~/.Bella/datasets/.` directory within your user space under the 
        dataset_name.

        To split the bella.data_types.TargetCollection.split_dataset method is 
        used.

        :param dataset_name: Name to associate to the dataset e.g. 
                             `SemEval 2014 rest train`. If split is not None 
                             then use a List of Strings e.g. 
                             [`SemEval 2014 rest train`, 
                             `SemEval 2014 rest dev`]
        :param split: Whether or not to split the dataset into train, test 
                      split. If not use None else specify the fraction of 
                      the data to use for the test split.
        :param cache: If the data is already saved use the Cache. Default 
                      is to use the cached data.
        :param group_by_sentence: Whether the data should be grouped by sentence
                                  this will then produce json lines that can 
                                  contain more than one target in a sentence.
        :param random: Whether the splitting of the training and test should 
                       be random or not.
        '''
        def create_json_file(fp: Path, data: List[Dict[str, Any]]) -> None:
            '''
            Given the a list of dictionaries that represent the Target data 
            converts these samples into json encoded samples which are saved on 
            each line within the file at the given file path(fp)

            :param fp: File path that will store the json samples one per line
            :param data: List of dictionaries that represent the Target data.
            :return: Nothing that data will be saved to the file.
            '''                
            with fp.open('w+') as json_file:
                if group_by_sentence:
                    data = TargetCollection([Target(**d) for d in data])
                    data = data.grouped_sentences
                    for index, target_datas in enumerate(data.values()):
                        text = target_datas[0]['text']
                        sentiments = []
                        targets = []
                        spans = []
                        for target_data in target_datas:
                            sentiments.append(target_data['sentiment'])
                            spans.append(target_data['spans'])
                            targets.append(target_data['target'])
                        sentence_data = {'text': text, 'sentiments': sentiments,
                                         'targets': targets, 'spans': spans}
                        json_encoded_data = json.dumps(sentence_data)
                        if index != 0:
                            json_encoded_data = f'\n{json_encoded_data}'
                        json_file.write(json_encoded_data)
                else:
                    for index, target_data in enumerate(data):
                        json_encoded_data = json.dumps(target_data)
                        if index != 0:
                            json_encoded_data = f'\n{json_encoded_data}'
                        json_file.write(json_encoded_data)
            
        # If splitting the data there has to be two dataset names else one name
        if split is None:
            assert isinstance(dataset_name, str)
        elif isinstance(split, float):
            assert isinstance(dataset_name, list)
            assert len(dataset_name) == 2
        
        BELLA_DATASET_DIR.mkdir(parents=True, exist_ok=True)

        dataset_names = dataset_name
        if not isinstance(dataset_name, list):
            dataset_names = [dataset_name]
        all_paths_exist = True
        dataset_paths = []
        for name in dataset_names:
            dataset_path = BELLA_DATASET_DIR.joinpath(name)
            if not dataset_path.exists():
                all_paths_exist = False
            dataset_paths.append(dataset_path)
        # Caching
        if cache and all_paths_exist:
            print(f'Using cache for the follwoing datasets: {dataset_names}')
            if split is None:
                return dataset_paths[0]
            return dataset_paths

        if split is None:
            create_json_file(dataset_paths[0], self.data_dict())
            return dataset_paths[0]
        # Splitting
        train, test = self.split_dataset(self, test_split=split, 
                                         random=random)
        create_json_file(dataset_paths[0], train.data_dict())
        create_json_file(dataset_paths[1], test.data_dict())
        return dataset_paths

    def categories_targets(self, filter: int = 2, coarse: bool = False
                           ) -> Tuple[Dict[str, List[str]], 
                                      Dict[str, List[str]]]:
        '''
        Returns two dictionaries. The first is a category to target list and 
        the second is a target to category list. Their is the option to have a 
        coarse grained cateogires option.
        '''
        if 'category' not in self.data()[0]:
            raise ValueError('The current TargetCollection must contain '
                             'targets that have a category key')

        temp_category_targets = defaultdict(set)
        for data in self.data():
            category = data['category']
            if coarse:
                category = category.split('#')[0]
            target = data['target']
            temp_category_targets[category].add(target)
        
        category_targets = {}
        target_categories = defaultdict(list)
        categories_filtered = []
        for category, targets in temp_category_targets.items():
            if len(targets) < filter:
                categories_filtered.append(category)
                continue
            category_targets[category] = list(targets)
            for target in targets:
                target_categories[target].append(category)
        # Which categories are filtered
        print(f'Filtered {len(categories_filtered)} categories which are:')
        for category in categories_filtered:
            print(category)

        return category_targets, dict(target_categories)

    @staticmethod
    def target_targets(target_set: set, 
                       category_targets: Dict[str, List[str]]
                       ) -> Dict[str, List[str]]:
        '''
        Returns a dictionary of targets as keys and values are a list of 
        realted targets.

        The related targets have all come from category_targets dictionary. 
        Where a target is related if it is in the same category. If it is in 
        multiple categories then all targets from each of those categories 
        are related.

        :param target_set: A list of targets that are candiate keys in the 
                           returned dictionary.
        :param category_targets: A dictionary of latent categories as keys 
                                 and related targets as values.
        :returns: Dictionary of targets and their related targets as values 
                  where the related targets have come from the category_targets
                  dictionary.
        '''
        target_rel_targets = {}
        for target in target_set:
            rel_targets = []
            for _, cat_targets in category_targets.items():
                if target in cat_targets or target.lower() in cat_targets:
                    rel_targets.extend(cat_targets)
                else:
                    lower_cat_targets = [cat_target.lower() 
                                         for cat_target in cat_targets]
                    if target.lower() in lower_cat_targets:
                        rel_targets.extend(cat_targets)
            if rel_targets:
                rel_targets = list(set(rel_targets))

                to_remove = set()
                for rel_target in rel_targets:
                    temp_rel_target = rel_target.lower()
                    if target.lower() == temp_rel_target:
                        to_remove.add(rel_target)
                for target_to_remove in to_remove:
                    rel_targets.remove(target_to_remove)
                if not to_remove:
                    raise ValueError('The target that maps to the list of '
                                     'related target should be in the related '
                                     'targets before being filtered out')
                target_rel_targets[target] = rel_targets
        return target_rel_targets

    def target_set(self, lower: bool = False) -> set:
        '''
        Returns a set of all the targets within this dataset.

        :param lower: Whether to return the targets lower cased.
        :returns: A set of all the targets within this dataset.
        '''
        targets = set()
        for data in self.data():
            target = data['target']
            if lower:
                target = target.lower()
            targets.add(target)
        return targets

    def subset_by_ids(self, ids: List[str]) -> 'TargetCollection':
        '''
        Returns a TargetCollection which is a subset of the current one. 
        Where the samples included in the subset are those with a `target_id`
        that is within the ids List.

        :param ids: A list of `target_id`s that are to be included in the 
                    subset
        :returns: A subset of the current TargetCollection where the subset 
                  only includes samples that have a `target_id` in the ids 
                  given.
        '''
        subset_data = []
        for data_id in ids:
            rel_data = dict(self[data_id].items())
            rel_data['target_id'] = data_id
            subset_data.append(Target(**rel_data))
        return TargetCollection(subset_data)

    def dataset_metric_scores(self, 
                              metric: Callable[[np.ndarray, np.ndarray], 
                                               float], 
                              **metric_kwargs) -> np.ndarray:
        '''
        Given a metric like accuracy returns all of the metric scores for the 
        dataset. This assumes that the dataset has the `predicted` sentiment 
        slot assigned.

        :param metric: Metric function e.g. f1_score
        :param metric_kwargs: Keyword arguments to provide to the metric 
                              function e.g. `average` = `macro`
        :returns: An array of metric results. One for each column in the 
                  predicted sentiment array.
        '''
        true_labels = self.sentiment_data()
        pred_matrix = self.sentiment_data(sentiment_field='predicted')
        score_vector = np.apply_along_axis(metric, 0, pred_matrix, 
                                           true_labels, **metric_kwargs)
        return score_vector


    # Not tested
    def targets_per_sentence(self):
        '''
        :returns: Dictionary of number of targets as keys and values the number \
        of sentences that have that many targets per sentence.
        :rtype: dict

        :Example:
        If we have 5 sentences that contain 1 target each and 4 sentences that
        contain 3 targets each then it will return a dict like:
        {1 : 5, 3 : 4}
        '''

        targets_sentence = {}
        for targets in self.grouped_sentences.values():
            num_targets = len(targets)
            targets_sentence[num_targets] = targets_sentence.get(num_targets, 0) + 1
        return targets_sentence

    # Not tested
    def avg_targets_per_sentence(self):
        return len(self) / self.number_sentences()
    # Not tested
    def number_sentences(self):
        return len(self.grouped_sentences)
    # Not tested
    def number_unique_targets(self):
        target_count = {}
        for target_instance in self.values():
            target = target_instance['target']
            target_count[target] = target_count.get(target, 0) + 1
        return len(target_count)

    def no_targets_sentiment(self):
        sentiment_targets = {}
        for target in self.values():
            sentiment = target['sentiment']
            sentiment_targets[sentiment] = sentiment_targets.get(sentiment, 0) + 1
        return sentiment_targets

    def ratio_targets_sentiment(self):
        no_sentiment_target = self.no_targets_sentiment()
        total_targets = sum(no_sentiment_target.values())
        ratio_sentiment_targets = {}
        for sentiment, no_targets in no_sentiment_target.items():
            ratio_sentiment_targets[sentiment] = round(no_targets / total_targets, 2)
        return ratio_sentiment_targets

    @property
    def grouped_sentences(self) -> Dict[str, List['Target']]:
        '''
        A dictionary of sentence_id as keys and a list of target instances that 
        have the same sentence_id as values.

        It stores a cache of this result and the cache will expire once the 
        data has changed within itself and then this value will have to be 
        recomputed.

        :returns: A dictionary of sentence_id as keys and a list of target 
                  instances that have the same sentence_id as values.
        '''

        # If the data has changed re-compute or if the data has never been 
        # compute, compute else return the cached results.
        if self._data_has_changed or self._grouped_sentences is None:
            self._data_has_changed = False
            sentence_targets = defaultdict(list)
            for target in self.data():
                if 'sentence_id' not in target:
                    raise ValueError(f'A Target id instance {target} does not '
                                     'have a sentence_id which is required.')
                sentence_id = target['sentence_id']
                sentence_targets[sentence_id].append(target)
            self._grouped_sentences = sentence_targets
            
        return self._grouped_sentences

    @property
    def grouped_sentiments(self) -> Dict[Any, List['Target']]:
        '''
        A dictionary where the keys are all possible sentiment values, the 
        value is a list of Target(s) that only have that associated sentiment 
        value.

        It stores a cache of this result and the cache will expire once the 
        data has changed within itself and then this value will have to be 
        recomputed.

        :returns: A dictionary of sentiment values as keys and a list of target 
                  instances that have the same sentiment value.
        '''

        # If the data has changed re-compute or if the data has never been 
        # compute, compute else return the cached results.
        if self._data_has_changed_gs or self._grouped_sentiments is None:
            self._data_has_changed_gs = False
            sentiment_targets = defaultdict(list)
            for target in self.data():
                sentiment_targets[target['sentiment']].append(target)
            self._grouped_sentiments = sentiment_targets
            
        return self._grouped_sentiments  

    @property
    def grouped_distinct_sentiments(self) -> Dict[Any, List['Target']]:
        '''
        A dictionary where the keys are the number of distinct sentiments that 
        the targets have in the associated value. A target has two distinct 
        sentiments if the associated sentence has contains targets that take 
        on one of two sentiment where at least one target has a different 
        sentiment to the rest. The targets in the values of this dictionary 
        are stored as a list of Target

        It stores a cache of this result and the cache will expire once the 
        data has changed within itself and then this value will have to be 
        recomputed.

        :returns: A dictionary of distinct sentiments per sentence as keys and 
                  a list of target instances that have the same distinct 
                  sentiments per sentence
        '''

        # If the data has changed re-compute or if the data has never been 
        # compute, compute else return the cached results.
        if self._data_has_changed_gds or self._grouped_distinct_sentiments is None:
            self._data_has_changed_gds = False
            distinct_sentiment_targets = defaultdict(list)
            for targets in self.grouped_sentences.values():
                target_col = TargetCollection(targets)
                num_unique_sentiments = len(target_col.stored_sentiments())
                distinct_sentiment_targets[num_unique_sentiments].extend(targets)
            self._grouped_distinct_sentiments = distinct_sentiment_targets
        return self._grouped_distinct_sentiments

    def group_by_sentence(self):
        '''
        This is now deprecated, please use grouped_sentences property.

        :returns: A dictionary of sentence_id as keys and a list of target \
        instances that have the same sentence_id as values.
        :rtype: defaultdict (default is list)
        '''

        dep_warning = ('This is now deprecated, please use grouped_sentences '
                       'property.')
        warnings.warn(dep_warning, DeprecationWarning)

        sentence_targets = defaultdict(list)
        for target in self.data():
            if 'sentence_id' not in target:
                raise ValueError('A Target id instance {} does not have '\
                                 'a sentence_id which is required.'\
                                 .format(target))
            sentence_id = target['sentence_id']
            sentence_targets[sentence_id].append(target)
        return sentence_targets

    def avg_constituency_depth(self):
        avg_depths = []
        for data in self.values():
            sentence_trees = constituency_parse(data['text'])
            tree_depths = [tree.height() - 1 for tree in sentence_trees]
            avg_depth = sum(tree_depths) / len(sentence_trees)
            avg_depths.append(avg_depth)
        return sum(avg_depths) / len(avg_depths)

    def avg_sentence_length_per_target(self, tokeniser=whitespace):
        all_sentence_lengths = []
        for data in self.values():
            all_sentence_lengths.append(len(tokeniser(data['text'])))
        return sum(all_sentence_lengths) / len(all_sentence_lengths)

    def word_list(self, tokeniser: Callable[[str], List[str]],
                  min_df: int = 0, lower: bool = True) -> List[str]:
        '''
        :param tokeniser: Tokeniser function to tokenise the text 
                          of each target/sample
        :param min_df: Optional. The minimum percentage of documents a 
                       token must occur in.
        :param lower: Optional. Whether to lower the text. 
        :return: A word list of all tokens that occur in this data collection 
                 given min_df.
        '''

        token_df = defaultdict(lambda: 0)
        num_df = 0
        for target in self.values():
            num_df += 1
            tokens = tokeniser(target['text'])
            for token in tokens:
                if lower:
                    token = token.lower()
                token_df[token] += 1
        min_df_value = int((num_df / 100) * min_df)
        word_list = [token for token, df in token_df.items()
                     if df > min_df_value]
        return word_list

    @staticmethod
    def combine_collections(*args):
        all_targets = []
        for collections in args:
            all_targets.extend(collections.data())
        return TargetCollection(all_targets)

    def __eq__(self, other):

        if len(self) != len(other):
            return False
        for key in self:
            if key not in other:
                return False
        return True

    def __repr__(self):
        '''
        :returns: String returned is what user see when the instance is \
        printed or printed within a interpreter.
        :rtype: String
        '''

        target_strings = ''

        self_len = len(self)
        if self_len > 2:
            for index, target in enumerate(self.data()):
                if index == 0:
                    target_strings += '{} ... '.format(target)
                if index == self_len - 1:
                    target_strings += '{}'.format(target)
        else:
            for target in self.data():
                target_strings += '{}, '.format(target)
        if target_strings != '':
            target_strings = target_strings.rstrip(', ')
        return 'TargetCollection({})'.format(target_strings)



'''
Contains functions that perform depdency parsing.
'''
from collections import defaultdict
from pathlib import Path
from typing import List, Tuple

from filelock import FileLock
import networkx as nx
from ruamel.yaml import YAML
from tweebo_parser import API

from bella.dependency_tokens import DependencyToken
from bella import stanford_tools

BELLA_CONFIG_FP = Path.home().joinpath('.Bella', 'config.yaml')
BELLA_CONFIG_LOCK_FP = Path.home().joinpath('.Bella', 'config.yaml.lock')

class TweeboParser(object):
    '''
    Singleton Class instance
    '''

    instance = None

    @staticmethod
    def get_config() -> Tuple[str, int]:
        hostname = '0.0.0.0'
        port = 8000
        yaml = YAML()
        config_data = {}
        BELLA_CONFIG_FP.parent.mkdir(parents=True, exist_ok=True)

        lock = FileLock(BELLA_CONFIG_LOCK_FP)
        with lock.acquire(60):
            if BELLA_CONFIG_FP.exists():
                with BELLA_CONFIG_FP.open('r') as config_file:
                    config_data = yaml.load(config_file)
                    if 'tweebo_parser' in config_data:
                        tweebo_config = config_data['tweebo_parser']
                        if 'hostname' in tweebo_config:
                            hostname = tweebo_config['hostname']
                        if 'port' in tweebo_config:
                            port = tweebo_config['port']

            config_data['tweebo_parser'] = {}
            config_data['tweebo_parser']['hostname'] = hostname
            config_data['tweebo_parser']['port'] = port
            with BELLA_CONFIG_FP.open('w') as config_file:
                yaml.dump(config_data, config_file)
        return hostname, port

    def __new__(cls):
        if TweeboParser.instance is None:
            hostname, port = cls.get_config()
            TweeboParser.instance = API(hostname=hostname, port=port, 
                                        log_errors=True)
        return TweeboParser.instance

    def __getattr__(self, name):
        return getattr(self.instance, name)

    def __setattr__(self, name, value):
        return setattr(self.instance, name, value)


def _to_dependencies_tokens(token_dep_sentence: List[Tuple[str, int]]
                            ) -> List[DependencyToken]:
    '''
    NOTE:
    The DependencyToken allows easy access to all the dependency links for that
    token.

    :param token_dep_sentence: list of tuples that contain (token, linked \
    index token)
    :type token_dep_sentence: list
    :returns: A list of DependencyToken instances one for each tuple/token.
    :rtype: list
    '''

    def dep_search(dep_index, sentence, dep_info):
        '''
        This is a tail recusive function that returns a dictionary denoting
        which index in the sentence relates to the other tokens in the sentence
        and at what dependency depth.

        :param dep_index: The index of the token whos dependecies are being \
        collected
        :param sentence: List of tuples which contain (token, index of \
        head word)
        :param dep_info: default dict whose keys are indexs from the sentence \
        and value is a default dict whose keys are dependency depths and \
        value is the sentence index related to that dependency depth.
        :type dep_index: int
        :type sentence: list
        :type dep_info: defaultdict
        :returns: Default dictionary whose keys are sentence indexs, values \
        are default dictionaries whose keys are dependency depths and value \
        is the associated sentence index to that depth.
        :rtype: defaultdict

        :Example:
        >>> sentence = [('To', -1), ('appear', 0), ('(', -2),
                        ('EMNLP', 1), ('2014', 3)]
        >>> dep_info = defaultdict(lambda: dict())
        >>> print(dep_search(4, sentence, dep_info))
        >>> {0 : {1 : 1, 2 : 3, 3 : 4},
             1 : {1 : 3, 2 : 3},
             3 : {1 : 4},
             4 : {}}
        '''

        head_index = sentence[dep_index][1]
        if head_index == -1 or head_index == -2:
            return dep_info
        prev_dep_info = dep_info[dep_index]
        head_deps = {dep_level + 1: deps
                     for dep_level, deps in prev_dep_info.items()}
        head_deps[1] = dep_index
        dep_info[head_index] = head_deps
        return dep_search(head_index, sentence, dep_info)

    dep_results = defaultdict(lambda: defaultdict(set))
    for index, _ in enumerate(token_dep_sentence):
        dep_result = dep_search(index, token_dep_sentence,
                                defaultdict(lambda: dict()))
        for token_index, dependencies in dep_result.items():
            for dep_level, dependent in dependencies.items():
                dep_results[token_index][dep_level].add(dependent)
    G = nx.Graph()
    for index, token_related_index in enumerate(token_dep_sentence):
        G.add_node(index)
        token, related_index = token_related_index
        related_index = related_index
        if related_index not in {-2, -1}:
            G.add_edge(index, related_index)

    # Convert each of the tokens in the sentence into a dependent token
    # using the results from searching through the dependencies
    dep_tokens = []
    for token_index, token_dep in enumerate(token_dep_sentence):
        token, _ = token_dep
        depth_related = dep_results[token_index]
        token_relations = defaultdict(list)

        connected_indexs = set()
        for _, node_relations in nx.bfs_successors(G, token_index):
            for node_relation in node_relations:
                connected_indexs.add(node_relation)
        if token_index not in connected_indexs:
            connected_indexs.add(token_index)
        connected_indexs = sorted(list(connected_indexs))
        connected_words = []
        for connected_index in connected_indexs:
            connected_word = token_dep_sentence[connected_index][0].strip()
            if connected_index == token_index:
                connected_words.append((connected_word, 'CURRENT'))
            else:
                connected_words.append((connected_word, 'CONNECTED'))
        # Get the tokens relations
        for depth, related_tokens_index in depth_related.items():
            for related_token_index in related_tokens_index:
                related_token = token_dep_sentence[related_token_index][0]
                token_relations[depth].append(related_token)
        dep_tokens.append(DependencyToken(token, token_relations,
                                          connected_words))
    return dep_tokens


def tweebo(texts: List[str]) -> List[DependencyToken]:
    '''
    Given a list of Strings will tokenise, pos tag and then dependecy parse
    the text using `Tweebo <https://github.com/ikekonglp/TweeboParser>`_
    a Tweet specific parser.

    The Tweebo parser cannot handle no strings therefore a special empty string
    symbol is required.

    If one of the texts is an empty String then an empty list will be returned
    for that index of the returned list.

    :param texts: The texts that are to be parsed
    :type text: list
    :returns: A list of of a list of DependencyToken instances. A list per \
    text in the texts argument.
    :rtype: list
    '''

    tweebo_api = TweeboParser()
    texts = [text.replace('\n', ' ') for text in texts]
    processed_texts = tweebo_api.parse_conll(texts)

    dep_texts = []
    for processed_text in processed_texts:
        token_dep_indexs = _convert_conll(processed_text)
        dep_texts.append(_to_dependencies_tokens(token_dep_indexs))
    return dep_texts


def _convert_conll(conll_data: str) -> List[Tuple[str, int]]:
    token_dep_indexs = []
    for line in conll_data.split('\n'):
        if not line:
            continue
        column_data = line.split('\t')
        word = column_data[1].strip()
        # All the word indexs start at 1 not 0.
        # Need to take into account the previous sentence words
        dep_token_index = int(column_data[6]) - 1
        token_dep_indexs.append((word, dep_token_index))
    return token_dep_indexs


def stanford(texts: str) -> List[DependencyToken]:

    dep_texts = []
    for text in texts:
        dep_dicts, tokens_dicts = stanford_tools.dependency_parse(text)
        token_dep_indexs = []
        prev_sent_length = 0
        for sentence, _ in enumerate(tokens_dicts):
            tokens_dict = tokens_dicts[sentence]
            dep_dict = dep_dicts[sentence]
            for i in range(1, len(tokens_dict) + 1):
                word = tokens_dict[i]['word']
                # All the word indexs start at 1 not 0.
                # Need to take into account the previous sentence words
                dep_word_index = (dep_dict[i][1] - 1) + prev_sent_length
                # If True then it is the root word
                if dep_word_index + 1 == prev_sent_length:
                    dep_word_index = -1
                token_dep_indexs.append((word, dep_word_index))
            prev_sent_length += len(tokens_dict)
        dep_texts.append(_to_dependencies_tokens(token_dep_indexs))
    return dep_texts



'''
Functions that allow evaluating model predictions easier, as well as custom
evaluation metrics specific for Target Based Sentiment Analysis e.g. the
:py:func:`bella.evaluation.distinct_sentiment_metrics` can return the
accuracy of the models predictions on samples that have a certain number of
distinct sentiments.

Functions:

1. score -- Given a metric, test data and predictions will evaluate the
   predictions based on the given metric.
2. evaluate_model -- Given a metric, dictionary of dataset names with test data
   and predictions will evaluate the predictions for each dataset.
3. evaluate_models -- Given a metric, dictionary of dataset names with test
   data and a dictionary of models and their predictions will evaluate the
   predictions for each dataset and for each model.
4. distinct_sentiment_metrics -- Custom metric function. It performs the
   `metric_func` on a subset of the test data where the samples have to come
   from sentences that contain a certain number of distinct_sentiments.
'''
from collections import defaultdict
import copy
from itertools import product
from typing import Callable, Union, Dict, List, Tuple, Any, Optional

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

import bella
from bella.data_types import TargetCollection


def score(metric: Union[Callable[[np.ndarray, np.ndarray], float],
                        Callable[['bella.data_types.TargetCollection',
                                  np.ndarray], float]],
          test_data: 'bella.data_types.TargetCollection',
          predictions: np.ndarray,
          custom_metric: bool = False, **metric_kwargs) -> float:
    '''
    Given a metric, test data and predictions will evaluate the predictions
    based on the given metric.

    :param metric: Metric used to evaluate the predictions e.g.
                   :py:func:`sklearn.metrics.f1_score` or a custom function
                   like :py:func:`bella.evaluation.distinct_sentiment_metrics`
    :param test_data: The test data stored as a
                      :py:class:`bella.data_types.TargetCollection` object.
    :param predictions: Array of predictions for the test data
    :param custom_metric: If the metric being used has come from the Bella
                          package then it is custom and this should be True
                          else use the Default False.
    :param metric_kwargs: Keyword arguments to the metric function.
    :return: The value given back by the metric function.
    '''
    if custom_metric:
        return metric(test_data, predictions, **metric_kwargs)
    return metric(test_data.sentiment_data(), predictions, **metric_kwargs)


def evaluate_model(metric: Union[Callable[[np.ndarray, np.ndarray], float],
                                 Callable[['bella.data_types.TargetCollection',
                                           np.ndarray], float]],
                   dataset_name_data: Dict[str, 'bella.data_types.TargetCollection'],
                   dataset_predictions: Dict[str, np.ndarray],
                   custom_metric: bool = False,
                   **metric_kwargs) -> Dict[str, float]:
    '''
    Given a metric, dictionary of dataset names with test data and predictions
    will evaluate the predictions for each dataset.

    :param metric: Metric used to evaluate the predictions e.g.
                   :py:func:`sklearn.metrics.f1_score` or a custom function
                   like :py:func:`bella.evaluation.distinct_sentiment_metrics`
    :param dataset_name_data: Dictionary of dataset names as keys and the
                              test data as values.
    :param dataset_predictions: Dictionary of dataset names as keys and the
                                predictions as values.
    :param custom_metric: If the metric being used has come from the Bella
                          package then it is custom and this should be True
                          else use the Default False.
    :param metric_kwargs: Keyword arguments to the metric function.
    :return: Dictionary of dataset names as keys and the metric score as
             values.
    '''
    dataset_score = {}
    for dataset_name, test_data in dataset_name_data.items():
        predictions = dataset_predictions[dataset_name]
        dataset_score[dataset_name] = score(metric, test_data, predictions,
                                            custom_metric, **metric_kwargs)
    return dataset_score


def evaluate_models(metric: Union[Callable[[np.ndarray, np.ndarray], float],
                                  Callable[['bella.data_types.TargetCollection',
                                            np.ndarray], float]],
                    dataset_name_data: Dict[str, 'bella.data_types.TargetCollection'],
                    model_dataset_predictions: Dict[str,
                                                    Dict[str, np.ndarray]],
                    custom_metric: bool = False, dataframe: bool = False,
                    **metric_kwargs) -> Union[Dict[str, Dict[str, float]],
                                              'pd.DataFrame']:
    '''
    Given a metric, dictionary of dataset names with test data and
    a dictionary of models and their predictions will evaluate the predictions
    for each dataset and for each model.

    :param metric: Metric used to evaluate the predictions e.g.
                   :py:func:`sklearn.metrics.f1_score` or a custom function
                   like :py:func:`bella.evaluation.distinct_sentiment_metrics`
    :param dataset_name_data: Dictionary of dataset names as keys and the
                              test data as values.
    :param model_dataset_predictions: Dictionary of dictionaries, outer
                                      dictionary is model name and inner
                                      dictionary is dataset names as keys and
                                      predictions as values.
    :param custom_metric: If the metric being used has come from the Bella
                          package then it is custom and this should be True
                          else use the Default False.
    :param dataframe: Whether to return the results as a
                      :py:class:`pandas.DataFrame`. Default False.
    :param metric_kwargs: Keyword arguments to the metric function.
    :return: Dictionary of dictionaries, outer dictionary is the model names,
             the inner is dataset names as keys and predictions as values. If
             `dataframe` argument is True then returns it as a
             :py:class:`pandas.DataFrame` object.
    '''
    model_dataset_score = defaultdict(lambda: dict())
    for model_name, dataset_predictions in model_dataset_predictions.items():
        for dataset_name, predictions in dataset_predictions.items():
            test_data = dataset_name_data[dataset_name]
            result = score(metric, test_data, predictions,
                           custom_metric, **metric_kwargs)
            model_dataset_score[model_name][dataset_name] = result
    if dataframe:
        model_dataset_score = pd.DataFrame(model_dataset_score)
        mean_values = model_dataset_score.mean()
        mean_df = mean_values.to_frame('Mean').T
        return pd.concat((model_dataset_score, mean_df))
    return model_dataset_score


def distinct_sentiment_metrics(test_data: 'bella.data_types.TargetCollection',
                               predictions: np.ndarray,
                               distinct_sentiments: int,
                               metric_func: Callable[[np.ndarray, np.ndarray],
                                                     float],
                               **metric_kwargs) -> float:
    '''
    Custom metric function. It performs the `metric_func` on a subset of the
    test data where the samples have to come from sentences that contain
    a certain number of distinct_sentiments.

    If a sentence has two samples where each target sample is positive this
    would mean that sentence only has one distinct_sentiment. If those two
    samples contained different sentiments then that sentence has two
    distinct_sentiments. This function evaluates predictions on samples that
    have come from sentences with a defined number of distinct_sentiments
    where the evaluation function is the `metric_func` argument over the
    subset of samples based on distinct_sentiments.

    :param test_data: The test data stored as a
                      :py:class:`bella.data_types.TargetCollection` object.
    :param predictions: Array of predictions for the test data
    :param distinct_sentiments: The number of unique sentiments in a sentence.
    :param metric_func: Metric used to evaluate the predictions e.g.
                        :py:func:`sklearn.metrics.f1_score`
    :param metric_kwargs: Keyword arguments to the metric function.
    '''
    test_data_copy = copy.deepcopy(test_data)
    test_data_copy.add_pred_sentiment(predictions)
    test_subset = test_data_copy.subset_by_sentiment(distinct_sentiments)
    true_values = test_subset.sentiment_data()
    subset_predictions = test_subset.sentiment_data(sentiment_field='predicted')
    return metric_func(true_values, subset_predictions, **metric_kwargs)

def get_kwargs(key: str, 
               potential_kwargs: Optional[Dict[str, Dict[str, Any]]] = None
               ) -> Dict[str, Any]:
    kwargs = {}
    if potential_kwargs is None:
        kwargs = {}
    elif key not in potential_kwargs:
        kwargs = {}
    else:
        kwargs = potential_kwargs[key]
    return kwargs

def plot_probability(data: pd.DataFrame, metric: str, 
                     model_name: Optional[str] = None,
                     models_to_remove: Optional[List[str]] = None,
                     bar_plot: bool = False, 
                     box_plot: bool = True,
                     cat_plot=False,
                     n_boot=100000,
                     **plot_kwargs):
    '''
    This assumes a dataframe that has the following columns:
    
    1. Metric
    2. Score
    3. Model
    4. Probability
    '''

    if model_name is not None:
        data = data[data['Model'] == model_name]
    data = data[data['Metric'] == metric]
    
    if not cat_plot:
        fig, ax = plt.subplots(1,1, **plot_kwargs)
    if box_plot:
        sns.boxplot(x='Probability', y='Score', hue='Model', 
                    data=data, ax=ax)
    elif bar_plot:
        if model_name is None:
            sns.barplot(x='Probability', y='Score', hue='Model', n_boot=n_boot,
                         data=data, ax=ax, ci=95)
        else:
            sns.barplot(x='Probability', y='Score', n_boot=n_boot,
                        data=data, ax=ax, ci=95)
    else:
        ax = sns.catplot(x='Probability', y='Score',  
                    col='Model', data=data, kind='bar',
                    n_boot=n_boot,
                    col_wrap=3, facet_kws=plot_kwargs)
        
    min_score = data['Score'].min()
    min_score = min_score - (min_score / 100)
    max_score = data['Score'].max()
    max_score = max_score + (max_score / 100)
    if not cat_plot:
        ax.set_ylim([min_score, max_score])
    else:
        ax.set(ylim=(min_score, max_score))
    return ax

def plot_acc_f1(results: pd.DataFrame, title: str, **plot_kwargs
                ) -> plt.Figure:
    fig, axs = plt.subplots(1, 2, **plot_kwargs)
    accuracy_data = results[results['Metric']=='Accuracy']
    f1_data = results[results['Metric']=='F1']
    axs[0] = sns.boxplot(x='Model', y='Score', data=accuracy_data, ax=axs[0])
    axs[0].set_ylabel('Accuracy')
    axs[0].set_title(title)
    axs[1] = sns.boxplot(x='Model', y='Score', data=f1_data, ax=axs[1])
    axs[1].set_ylabel('F1')
    axs[1].set_title(title)
    return fig

def datasets_df(datasets: TargetCollection, 
                metrics: List[Tuple[str, 
                                    Callable[[np.ndarray, np.ndarray], 
                                             float]]],
                metric_funcs_kwargs: Optional[Dict[str, Dict[str, Any]]] = None,
                additional_data: Optional[Dict[str, Dict[str, Any]]] = None
                ) -> pd.DataFrame:
    num_datasets = len(datasets)
    dataset_names = [dataset.name for dataset in datasets]
    if len(set(dataset_names)) != num_datasets:
        raise ValueError('datasets given must have unique name these datasets'
                         f' do not: {dataset_names}')
    df_data = defaultdict(list)
    for dataset in datasets:
        dataset_name = dataset.name
        for metric_name, metric_func in metrics:
            metric_kwargs = get_kwargs(metric_name, metric_funcs_kwargs)
            scores = dataset.dataset_metric_scores(metric_func, **metric_kwargs)
            for score in scores:
                df_data['Model'].append(dataset_name)
                df_data['Metric'].append(f'{metric_name}')
                df_data['Score'].append(score)
                if additional_data is None:
                    continue    
                if dataset_name in additional_data:
                    additional_cols = additional_data[dataset_name]
                    for column_name, value in additional_cols.items():
                        df_data[column_name].append(value)
    return pd.DataFrame(df_data)

def summary_errors(datasets: TargetCollection,
                   metrics: List[Tuple[str, 
                                       Callable[[np.ndarray, np.ndarray], 
                                                float]]],
                   error_funcs: List[Tuple[str, Callable[[TargetCollection], 
                                                         List[str]]]],
                   metric_funcs_kwargs: Optional[Dict[str, Dict[str, Any]]] = None,
                   error_funcs_kwargs: Optional[Dict[str, Dict[str, Any]]] = None,
                   std_err: bool = True
                   ) -> pd.DataFrame:

    num_datasets = len(datasets)
    dataset_names = [dataset.name for dataset in datasets]
    if len(set(dataset_names)) != num_datasets:
        raise ValueError('datasets given must have unique name these datasets'
                         f' do not: {dataset_names}')

    error_names = [error_func_name for error_func_name, _ in error_funcs]
    metric_names = [metric_name for metric_name, _ in metrics]
    columns = list(product(error_names, metric_names))
    num_columns = len(columns)
    columns = pd.MultiIndex.from_tuples(columns)
    summary_data = pd.DataFrame(np.empty((num_datasets, num_columns)),
                                index=dataset_names, columns=columns)

    error_name_ids = []
    for error_func_name, error_func in error_funcs:
        kwargs = get_kwargs(error_func_name, error_funcs_kwargs)
        error_ids = error_func(datasets[0], **kwargs)
        error_name_ids.append((error_func_name, error_ids))
    for error_name, error_ids in error_name_ids:
        for dataset in datasets:
            dataset_name = dataset.name
            error_data = dataset.subset_by_ids(error_ids)
            for metric_name, metric_func in metrics:
                metric_kwargs = get_kwargs(metric_name, metric_funcs_kwargs)
                scores = error_data.dataset_metric_scores(metric_func, 
                                                          **metric_kwargs)
                mean_score = np.mean(scores)
                score = mean_score * 100
                score = f'{score:.2f}'
                if std_err:
                    std_score = np.std(scores) * 100
                    score = f'{score} ({std_score:.2f})'
                summary_data.loc[(dataset_name), 
                                 (error_name, metric_name)] = score
    return summary_data


'''
SKLearn Transformer

Classes:

1. Debug -- Allows you to debug previous transformers
'''

import pdb

from sklearn.base import TransformerMixin
from sklearn.base import BaseEstimator

class Debug(BaseEstimator, TransformerMixin):
    '''
    SKLearn transformer that is to be used a debugger between different
    SKLearn transformers.

    Methods:

    1. fit - Does nothing as nothing is done at fit time.
    2. fit_transform - Performs the transform method.
    3. transform - Creates a REPL terminal to inspect the input transformed data
    '''
    def __init__(self):
        pass

    def fit(self, data, y=None):
        '''Kept for consistnecy with the TransformerMixin'''

        return self

    def fit_transform(self, data, y=None):
        '''see self.transform'''

        return self.transform(data)

    def transform(self, data):
        '''
        :param data: Data from the previous transformer in the pipeline
        :type data: array
        :returns: Nothing. Creates a Python interactive shell to inspect the \
        data from the previous transformation.
        :rtype: None
        '''

        pdb.set_trace()



from sklearn.base import TransformerMixin
from sklearn.base import BaseEstimator

from bella import lexicons

class LexiconFilter(BaseEstimator, TransformerMixin):

    def __init__(self, lexicon=None, zero_token='$$$ZERO_TOKEN$$$'):
        self.lexicon = lexicon
        self.zero_token = zero_token

    def fit(self, context_tokens, y=None):
        '''Kept for consistnecy with the TransformerMixin'''

        return self

    def fit_transform(self, context_tokens, y=None):
        '''see self.transform'''

        return self.transform(context_tokens)

    def transform(self, contexts_tokens):
        lexicon_words = self.lexicon.words
        context_tokens_filtered = []
        for context_tokens in contexts_tokens:
            all_tokens_filtered = []
            for context in context_tokens:
                context_tokens = []
                for token in context:
                    if token not in lexicon_words:
                        context_tokens.append(self.zero_token)
                    else:
                        context_tokens.append(token)
                all_tokens_filtered.append(context_tokens)
            context_tokens_filtered.append(all_tokens_filtered)
        return context_tokens_filtered



'''
This module contains the Lexicon classes where each class is a sub class of
the :py:class:`bella.lexicons.Lexicon` class.

The following classes are in the module:

1. Lexicon -- Base class
2. HuLiu -- `Hu and Liu <https://www.cs.uic.edu/~liub/publications/kdd04-revSummary.pdf>`_
lexicon.
3. Mpqa -- `Wilson, Wiebe and Hoffman \
<https://aclanthology.coli.uni-saarland.de/papers/H05-1044/h05-1044>`_ lexicon.
4. NRC -- `Mohammad and Turney \
<http://saifmohammad.com/WebDocs/Mohammad-Turney-NAACL10-EmotionWorkshop.pdf>`_
lexicon.
'''
from collections import defaultdict
import csv
import re
from pathlib import Path


class Lexicon():
    '''
    Base class for all lexicons.

    Attributes:

    1. self.lexicon -- List of tuples where each tuple contains (word, category) \
    e.g. [('great', 'positive')]
    3. self.words -- Set of Strings which are the words in the self.lexicon list.
    2. self.name -- String name of the class.

    Methods to be overidden:

    1. self.get_lexicon -- Must return List of tuples where each tuple contains \
    (word, category). This is not used to create self.lexicon if the lexicon \
    is defined in the constructor.

    Static Methods:

    1. combine_lexicons -- Combines two Lexicon classes together and removes \
    (word, category) tuples from self.lexicon that have different category values \
    for the same word.
    '''

    def __init__(self, subset_cats=None, lower=False, name=None, lexicon=None):
        if subset_cats is not None:
            if not isinstance(subset_cats, set):
                raise TypeError('subset_cats parameter has to be of type set '\
                                'and not {}'.format(type(subset_cats)))
        if lower != False:
            if not isinstance(lower, bool):
                raise TypeError('lower parameter has to be of type bool not {}'\
                                .format(type(lower)))
        self.lexicon = lexicon
        if self.lexicon is None:
            self.lexicon = self.get_lexicon()

        if not isinstance(self.lexicon, list):
            raise TypeError('self.lexicon has to be of type list not {} This could'\
                            ' be due to the `get_lexicon` method'\
                             .format(type(self.lexicon)))
        if not isinstance(self.lexicon[0], tuple):
            raise TypeError('self.lexicon should contain tuples not {} This could'\
                            ' be due to the `get_lexicon` method'\
                            .format(type(self.lexicon[0])))

        self.lexicon = self._process_lexicon(subset_cats, lower)
        self.remove_duplicates()
        self.words = set([word for word, cat in self.lexicon])
        if name is not None:
            self.name = name
        else:
            self.name = self.__class__.__name__

    def remove_duplicates(self):
        duplicate_lexicon = defaultdict(list)
        for word, cat in self.lexicon:
            duplicate_lexicon[word].append(cat)
        lexicon = []
        for word, cat_list in duplicate_lexicon.items():
            if len(cat_list) > 1:
                continue
            lexicon.append((word, cat_list[0]))
        self.lexicon = lexicon

    def get_lexicon(self):
        '''
        Method to be overidden. It should be the method that all sub classes
        use to parse in their lexicon into a list of tuples format where each
        tuple contain (word, category).

        :returns: A list of tuples containing (word, category) e.g. \
        ('great', 'positive')
        :rtype: list
        '''

        return [('example', 'value')]

    def _process_lexicon(self, subset_cats, lower):
        '''
        Method not to be overidden. It process the lexicon based on the parameters.
        Returns a list of tuples where a tuple contains (word, category).

        :param subset_cats: Categories of words that you want to return e.g. \
        `positive` for `positive` words only. If None then no words are subsetted.
        :param lower: Whether to lower case the words in the lexicon.
        :type subset_values: set
        :type lower: bool
        :returns: Returns the lexicon as a list of tuples (word, category).
        :rtype: list
        '''

        temp_lexicon = self.lexicon
        lexicon = set()
        for word, cat in temp_lexicon:
            if subset_cats is not None:
                if cat not in subset_cats:
                    continue
            if lower:
                word = word.lower()
            lexicon.add((word, cat))
        return list(lexicon)

    @staticmethod
    def combine_lexicons(lexicon1, lexicon2):
        '''
        Combines two lexicons and removes words that have different categories
        for the words in the lexicons. This can be used to combine sentiment
        lexicons and remove words that have opposite sentiments. The self.name
        of the Lexicon is the combination the names of lexicon1 and lexicon2.

        NOTE: Requires that both lexicons have the same categories.

        :param lexicon1: An instance of :py:class:`bella.lexicons.Lexicon`
        :param lexicon2: An instance of :py:class:`bella.lexicons.Lexicon`
        :type lexicon1: :py:class:`bella.lexicons.Lexicon`
        :type lexicon2: :py:class:`bella.lexicons.Lexicon`
        :returns: Combines the two lexicons and removes words that have opposite \
        sentiment and returns it as a :py:class:`bella.lexicons.Lexicon`
        :rtype: :py:class:`bella.lexicons.Lexicon`
        '''
        def compare_lexicons(lex1, lex2, lex_name):
            '''
            Given the two lexicons will combine the lexicons together and remove
            words that do not have the same category.

            :param lex1: An instance of :py:class:`bella.lexicons.Lexicon`
            :param lex2: An instance of :py:class:`bella.lexicons.Lexicon`
            :param lex_name: The name to give the returned Lexicon
            :type lex1: :py:class:`bella.lexicons.Lexicon`
            :type lex2: :py:class:`bella.lexicons.Lexicon`
            :type lex_name: String
            :returns: Combines the two lexicons and removes words that have opposite \
            sentiment and returns it as a :py:class:`bella.lexicons.Lexicon`
            :rtype: :py:class:`bella.lexicons.Lexicon`
            '''

            combined_words = set(list(lex1.keys()) + list(lex2.keys()))
            combined_lexicons = []
            for word in combined_words:

                if word in lex1 and word in lex2:
                    if lex1[word] != lex2[word]:
                        continue
                    combined_lexicons.append((word, lex1[word]))
                elif word in lex1:
                    combined_lexicons.append((word, lex1[word]))
                elif word in lex2:
                    combined_lexicons.append((word, lex2[word]))
                else:
                    raise KeyError('The word {} has to be in one of the lexicons'\
                                   .format(word))
            return Lexicon(lexicon=combined_lexicons, name=lex_name)

        if not isinstance(lexicon1, Lexicon) or not isinstance(lexicon2, Lexicon):
            raise TypeError('Both parameters require to be type Lexicon not {}, {}'\
                            .format(type(lexicon1), type(lexicon2)))
        combined_lex_name = "{} {}".format(lexicon1.name, lexicon2.name)
        lexicon1 = lexicon1.lexicon
        lexicon2 = lexicon2.lexicon

        word_cat1 = {word : cat for word, cat in lexicon1}
        word_cat2 = {word : cat for word, cat in lexicon2}

        cat1 = set(word_cat1.values())
        cat2 = set(word_cat2.values())
        if cat1 != cat2:
            raise ValueError('These two lexicons cannot be combined as they have '\
                             'different and non-comparable categories: '\
                             'categories1: {} categories2: {}'\
                             .format(cat1, cat2))
        return compare_lexicons(word_cat1, word_cat2, combined_lex_name)

    def __str__(self):
        return self.name


class Mpqa(Lexicon):
    '''
    MPQA lexicon `Wilson, Wiebe and Hoffman \
    <https://aclanthology.coli.uni-saarland.de/papers/H05-1044/h05-1044>`_.
    Sub class of :py:class:`bella.lexicons.Lexicon`

    Category lables = 1. positive, 2. negative, 3. both, 4. neutral
    '''

    def __init__(self, mpqa_fp: Path, subset_cats=None, lower=False,
                 name=None, lexicon=None):
        '''
        :param mpqa_fp: File path to the mpqa lexicon.
        '''
        self.file_path = mpqa_fp
        super().__init__(subset_cats=subset_cats, lower=lower,
                         name=name, lexicon=lexicon)

    def get_lexicon(self):
        '''
        Overrides :py:func@`bella.lexicons.Lexicon.get_lexicon`
        '''

        word_cats = []
        with self.file_path.open('r') as mpqa_file:
            for line in mpqa_file:
                line = line.strip()
                if line:
                    key_values = {}
                    for data in line.split():
                        if '=' in data:
                            key, value = data.split('=')
                            key_values[key] = value
                    word = key_values['word1']
                    cat = key_values['priorpolarity']
                    if cat == 'weakneg':
                        cat = key_values['polarity']
                    word_cats.append((word, cat))
        return word_cats


class HuLiu(Lexicon):
    '''
    `Hu and Liu
    <https://www.cs.uic.edu/~liub/publications/kdd04-revSummary.pdf>`_
    Lexicon.
    Sub class of :py:class:`bella.lexicons.Lexicon`

    Category lables = 1. positive, 2. negative
    '''

    def __init__(self, huliu_folder_path: Path, subset_cats=None, lower=False,
                 name=None, lexicon=None):
        '''
        :param nrc_fp: File path to the Hu Liu sentiment lexicon folder that \
        contains both the positive and negative words txt files.
        '''
        self.folder_path = huliu_folder_path
        super().__init__(subset_cats=subset_cats, lower=lower,
                         name=name, lexicon=lexicon)

    def get_lexicon(self):
        '''
        Overrides :py:func@`bella.lexicons.Lexicon.get_lexicon`
        '''

        cats = ['positive', 'negative']
        word_cat = []
        for cat in cats:
            file_path = self.folder_path.joinpath('{}-words.txt'.format(cat))
            with file_path.open('r', encoding='cp1252') as senti_file:
                for line in senti_file:
                    if re.search('^;', line) or re.search(r'^\W+', line):
                        continue
                    line = line.strip()
                    word_cat.append((line.strip(), cat))
        return word_cat


class NRC(Lexicon):
    '''
    NRC emotion `Mohammad and Turney \
    <http://saifmohammad.com/WebDocs/Mohammad-Turney-NAACL10-EmotionWorkshop.pdf>`_\
    Lexicon
    Sub class of :py:class:`bella.lexicons.Lexicon`

    Category lables = 1. anger, 2. fear, 3. anticipation, 4. trust, 5. surprise,
    6. sadness, 7. joy, 8. disgust, 9. positive, and 10. negative.
    '''

    def __init__(self, nrc_fp: Path, subset_cats=None, lower=False,
                 name=None, lexicon=None):
        '''
        :param nrc_fp: File path to the NRC lexicon.
        '''
        self.file_path = nrc_fp
        super().__init__(subset_cats=subset_cats, lower=lower,
                         name=name, lexicon=lexicon)

    def get_lexicon(self):
        '''
        Overrides :py:func:`bella.lexicons.Lexicon.get_lexicon`
        '''

        word_cat = []
        with self.file_path.open('r', newline='') as emotion_file:
            tsv_reader = csv.reader(emotion_file, delimiter='\t')
            for row in tsv_reader:
                if len(row):
                    word = row[0]
                    cat = row[1]
                    association = int(row[2])
                    if association:
                        word_cat.append((word, cat))
        return word_cat



'''
Module contains a Class that is a scikit learn Transformer.

Classes:

1. ContextTokeniser - Converts a list of String lists into token lists. See the
transformer method of the class for more details.
'''
from sklearn.base import TransformerMixin
from sklearn.base import BaseEstimator

from bella import tokenisers


class ContextTokeniser(BaseEstimator, TransformerMixin):
    '''
    Scikit learn transformer class. Converts list of String lists into tokens.

    Attributes:

    1. self.tokeniser - tokeniser function. Given a String returns a list of Strings.
    Default whitespace tokeniser.
    2. self.lower - whether to lower case the tokens. Default False.

    See :py:func:`bella.tokenisers` for more tokeniser functions that can be
    used here or create your own function.
    '''

    def __init__(self, tokeniser=tokenisers.whitespace, lower=False):
        self.tokeniser = tokeniser
        self.lower = lower

    def fit(self, target_contexts, y=None):
        '''Kept for consistnecy with the TransformerMixin'''

        return self

    def fit_transform(self, target_contexts, y=None):
        '''see self.transform'''

        return self.transform(target_contexts)

    def transform(self, target_contexts):
        '''
        Given a list of String lists where each String represents a context per
        target span it returns those Strings as a list of Strings (tokens).

        :param target_contexts: A list of String lists e.g. \
        [['It was nice this morning', 'It was nice this morning but not \
        yesterday morning'], ['another day']] where each String is a span context \
        for a target.
        :type target_contexts: list
        :returns: A list of Strings (tokens) per span context. e.g.
        [[['It', 'was', 'nice', 'this', 'morning'], ['It', 'was', 'nice', 'this',\
        'morning', 'but', 'not', 'yesterday', 'morning']], [['another', 'day']]]
        :rtype: list
        '''

        token_contexts = []
        for target_span_contexts in target_contexts:
            token_span_contexts = []
            for span_context in target_span_contexts:
                tokens = self.tokeniser(span_context)
                if self.lower:
                    tokens = [token.lower() for token in tokens]
                token_span_contexts.append(tokens)
            token_contexts.append(token_span_contexts)
        return token_contexts



# -*- coding: utf-8 -*-
#
# This file is part of moses.  Its use is licensed under the GNU Lesser General
# Public License version 2.1 or, at your option, any later version.

"""
Note: This Python tokenizer requires the following dependencies:

To Install
====

$ pip install -U six nltk
$ python -m nltk.downloader perluniprops
$ python -m nltk.downloader nonbreaking_prefixes

This has been taken from the `Moses project <https://github.com/moses-smt/mose\
sdecoder/blob/master/scripts/tokenizer/python-tokenizer/moses.py>`_
"""
import re

from six import text_type

from nltk.tokenize.api import TokenizerI
from nltk.tokenize.util import is_cjk
from nltk.corpus import perluniprops, nonbreaking_prefixes


class MosesTokenizer(TokenizerI):
    """
    This is a Python port of the Moses Tokenizer from
    https://github.com/moses-smt/mosesdecoder/blob/master/scripts/tokenizer/tokenizer.perl
    >>> tokenizer = MosesTokenizer()
    >>> text = u'This, is a sentence with weird\xbb symbols\u2026 appearing everywhere\xbf'
    >>> expected_tokenized = u'This , is a sentence with weird \xbb symbols \u2026 appearing everywhere \xbf'
    >>> tokenized_text = tokenizer.tokenize(text, return_str=True)
    >>> tokenized_text == expected_tokenized
    True
    >>> tokenizer.tokenize(text) == [u'This', u',', u'is', u'a', u'sentence', u'with', u'weird', u'\xbb', u'symbols', u'\u2026', u'appearing', u'everywhere', u'\xbf']
    True
    The nonbreaking prefixes should tokenize the final fullstop.
    >>> m = MosesTokenizer()
    >>> m.tokenize('abc def.')
    [u'abc', u'def', u'.']
    The nonbreaking prefixes should deal the situation when numeric only prefix is the last token.
    In below example, "pp" is the last element, and there is no digit after it.
    >>> m = MosesTokenizer()
    >>> m.tokenize('2016, pp.')
    [u'2016', u',', u'pp', u'.']
    
    >>> sent = "This ain't funny. It's actually hillarious, yet double Ls. | [] < > [ ] & You're gonna shake it off? Don't?"
    >>> m.tokenize(sent, escape=True)
    ['This', 'ain', '&apos;t', 'funny', '.', 'It', '&apos;s', 'actually', 'hillarious', ',', 'yet', 'double', 'Ls', '.', '&#124;', '&#91;', '&#93;', '&lt;', '&gt;', '&#91;', '&#93;', '&amp;', 'You', '&apos;re', 'gonna', 'shake', 'it', 'off', '?', 'Don', '&apos;t', '?']
    >>> m.tokenize(sent, escape=False)
    ['This', 'ain', "'t", 'funny', '.', 'It', "'s", 'actually', 'hillarious', ',', 'yet', 'double', 'Ls', '.', '|', '[', ']', '<', '>', '[', ']', '&', 'You', "'re", 'gonna', 'shake', 'it', 'off', '?', 'Don', "'t", '?']
    """

    # Perl Unicode Properties character sets.
    IsN = text_type(''.join(perluniprops.chars('IsN')))
    IsAlnum = text_type(''.join(perluniprops.chars('IsAlnum')))
    IsSc = text_type(''.join(perluniprops.chars('IsSc')))
    IsSo = text_type(''.join(perluniprops.chars('IsSo')))
    IsAlpha = text_type(''.join(perluniprops.chars('IsAlpha')))
    IsLower = text_type(''.join(perluniprops.chars('IsLower')))

    # Remove ASCII junk.
    DEDUPLICATE_SPACE = r'\s+', r' '
    ASCII_JUNK = r'[\000-\037]', r''

    # Neurotic Perl heading space, multi-space and trailing space chomp.
    # These regexes are kept for reference purposes and shouldn't be used!!
    MID_STRIP = r" +", r" "  # Use DEDUPLICATE_SPACE instead.
    LEFT_STRIP = r"^ ", r""  # Uses text.lstrip() instead.
    RIGHT_STRIP = r" $", r""  # Uses text.rstrip() instead.

    # Pad all "other" special characters not in IsAlnum.
    PAD_NOT_ISALNUM = u'([^{}\s\.\'\`\,\-])'.format(IsAlnum), r' \1 '

    # Splits all hypens (regardless of circumstances), e.g.
    # 'foo -- bar' -> 'foo @-@ @-@ bar' , 'foo-bar' -> 'foo @-@ bar'
    AGGRESSIVE_HYPHEN_SPLIT = u'([{alphanum}])\-(?=[{alphanum}])'.format(alphanum=IsAlnum), r'\1 \@-\@ '

    # Make multi-dots stay together.
    REPLACE_DOT_WITH_LITERALSTRING_1 = r'\.([\.]+)', ' DOTMULTI\1'
    REPLACE_DOT_WITH_LITERALSTRING_2 = r'DOTMULTI\.([^\.])', 'DOTDOTMULTI \1'
    REPLACE_DOT_WITH_LITERALSTRING_3 = r'DOTMULTI\.', 'DOTDOTMULTI'

    # Separate out "," except if within numbers (5,300)
    # e.g.  A,B,C,D,E > A , B,C , D,E
    # First application uses up B so rule can't see B,C
    # two-step version here may create extra spaces but these are removed later
    # will also space digit,letter or letter,digit forms (redundant with next section)
    COMMA_SEPARATE_1 = u'([^{}])[,]'.format(IsN), r'\1 , '
    COMMA_SEPARATE_2 = u'[,]([^{}])'.format(IsN), r' , \1'

    # Attempt to get correct directional quotes.
    DIRECTIONAL_QUOTE_1 = r'^``', r'`` '
    DIRECTIONAL_QUOTE_2 = r'^"', r'`` '
    DIRECTIONAL_QUOTE_3 = r'^`([^`])', r'` \1'
    DIRECTIONAL_QUOTE_4 = r"^'", r'`  '
    DIRECTIONAL_QUOTE_5 = r'([ ([{<])"', r'\1 `` '
    DIRECTIONAL_QUOTE_6 = r'([ ([{<])``', r'\1 `` '
    DIRECTIONAL_QUOTE_7 = r'([ ([{<])`([^`])', r'\1 ` \2'
    DIRECTIONAL_QUOTE_8 = r"([ ([{<])'", r'\1 ` '

    # Replace ... with _ELLIPSIS_
    REPLACE_ELLIPSIS = r'\.\.\.', r' _ELLIPSIS_ '
    # Restore _ELLIPSIS_ with ...
    RESTORE_ELLIPSIS = r'_ELLIPSIS_', r'\.\.\.'

    # Pad , with tailing space except if within numbers, e.g. 5,300
    # These are used in nltk.tokenize.moses.penn_tokenize()
    COMMA_1 = u'([^{numbers}])[,]([^{numbers}])'.format(numbers=IsN), r'\1 , \2'
    COMMA_2 = u'([{numbers}])[,]([^{numbers}])'.format(numbers=IsN), r'\1 , \2'
    COMMA_3 = u'([^{numbers}])[,]([{numbers}])'.format(numbers=IsN), r'\1 , \2'

    # Pad unicode symbols with spaces.
    SYMBOLS = u'([;:@#\$%&{}{}])'.format(IsSc, IsSo), r' \1 '

    # Separate out intra-token slashes.  PTB tokenization doesn't do this, so
    # the tokens should be merged prior to parsing with a PTB-trained parser.
    # e.g. "and/or" -> "and @/@ or"
    INTRATOKEN_SLASHES = u'([{alphanum}])\/([{alphanum}])'.format(alphanum=IsAlnum), r'$1 \@\/\@ $2'

    # Splits final period at end of string.
    FINAL_PERIOD = r"""([^.])([.])([\]\)}>"']*) ?$""", r'\1 \2\3'
    # Pad all question marks and exclamation marks with spaces.
    PAD_QUESTION_EXCLAMATION_MARK = r'([?!])', r' \1 '

    # Handles parentheses, brackets and converts them to PTB symbols.
    PAD_PARENTHESIS = r'([\]\[\(\){}<>])', r' \1 '
    CONVERT_PARENTHESIS_1 = r'\(', '-LRB-'
    CONVERT_PARENTHESIS_2 = r'\)', '-RRB-'
    CONVERT_PARENTHESIS_3 = r'\[', '-LSB-'
    CONVERT_PARENTHESIS_4 = r'\]', '-RSB-'
    CONVERT_PARENTHESIS_5 = r'\{', '-LCB-'
    CONVERT_PARENTHESIS_6 = r'\}', '-RCB-'

    # Pads double dashes with spaces.
    PAD_DOUBLE_DASHES = r'--', ' -- '

    # Adds spaces to start and end of string to simplify further regexps.
    PAD_START_OF_STR = r'^', ' '
    PAD_END_OF_STR = r'$', ' '

    # Converts double quotes to two single quotes and pad with spaces.
    CONVERT_DOUBLE_TO_SINGLE_QUOTES = r'"', " '' "
    # Handles single quote in possessives or close-single-quote.
    HANDLES_SINGLE_QUOTES = r"([^'])' ", r"\1 ' "

    # Pad apostrophe in possessive or close-single-quote.
    APOSTROPHE = r"([^'])'", r"\1 ' "

    # Prepend space on contraction apostrophe.
    CONTRACTION_1 = r"'([sSmMdD]) ", r" '\1 "
    CONTRACTION_2 = r"'ll ", r" 'll "
    CONTRACTION_3 = r"'re ", r" 're "
    CONTRACTION_4 = r"'ve ", r" 've "
    CONTRACTION_5 = r"n't ", r" n't "
    CONTRACTION_6 = r"'LL ", r" 'LL "
    CONTRACTION_7 = r"'RE ", r" 'RE "
    CONTRACTION_8 = r"'VE ", r" 'VE "
    CONTRACTION_9 = r"N'T ", r" N'T "

    # Informal Contractions.
    CONTRACTION_10 = r" ([Cc])annot ", r" \1an not "
    CONTRACTION_11 = r" ([Dd])'ye ", r" \1' ye "
    CONTRACTION_12 = r" ([Gg])imme ", r" \1im me "
    CONTRACTION_13 = r" ([Gg])onna ", r" \1on na "
    CONTRACTION_14 = r" ([Gg])otta ", r" \1ot ta "
    CONTRACTION_15 = r" ([Ll])emme ", r" \1em me "
    CONTRACTION_16 = r" ([Mm])ore$text =~ s='n ", r" \1ore 'n "
    CONTRACTION_17 = r" '([Tt])is ", r" '\1 is "
    CONTRACTION_18 = r" '([Tt])was ", r" '\1 was "
    CONTRACTION_19 = r" ([Ww])anna ", r" \1an na "

    # Clean out extra spaces
    CLEAN_EXTRA_SPACE_1 = r'  *', r' '
    CLEAN_EXTRA_SPACE_2 = r'^ *', r''
    CLEAN_EXTRA_SPACE_3 = r' *$', r''

    # Neurotic Perl regexes to escape special characters.
    # These XML escaping regexes are kept such that tokens generated from
    # NLTK's implementation is consistent with Moses' tokenizer's output.
    # Outside of the MosesTokenizer function, it's strongly encouraged to use
    # nltk.tokenize.util.xml_escape() function instead.
    ESCAPE_AMPERSAND = r'&', r'&amp;'
    ESCAPE_PIPE = r'\|', r'&#124;'
    ESCAPE_LEFT_ANGLE_BRACKET = r'<', r'&lt;'
    ESCAPE_RIGHT_ANGLE_BRACKET = r'>', r'&gt;'
    ESCAPE_SINGLE_QUOTE = r"\'", r"&apos;"
    ESCAPE_DOUBLE_QUOTE = r'\"', r'&quot;'
    ESCAPE_LEFT_SQUARE_BRACKET = r"\[", r"&#91;"
    ESCAPE_RIGHT_SQUARE_BRACKET = r"]", r"&#93;"

    EN_SPECIFIC_1 = u"([^{alpha}])[']([^{alpha}])".format(alpha=IsAlpha), r"\1 ' \2"
    EN_SPECIFIC_2 = u"([^{alpha}{isn}])[']([{alpha}])".format(alpha=IsAlpha, isn=IsN), r"\1 ' \2"
    EN_SPECIFIC_3 = u"([{alpha}])[']([^{alpha}])".format(alpha=IsAlpha), r"\1 ' \2"
    EN_SPECIFIC_4 = u"([{alpha}])[']([{alpha}])".format(alpha=IsAlpha), r"\1 '\2"
    EN_SPECIFIC_5 = u"([{isn}])[']([s])".format(isn=IsN), r"\1 '\2"

    ENGLISH_SPECIFIC_APOSTROPHE = [EN_SPECIFIC_1, EN_SPECIFIC_2, EN_SPECIFIC_3,
                                   EN_SPECIFIC_4, EN_SPECIFIC_5]

    FR_IT_SPECIFIC_1 = u"([^{alpha}])[']([^{alpha}])".format(alpha=IsAlpha), r"\1 ' \2"
    FR_IT_SPECIFIC_2 = u"([^{alpha}])[']([{alpha}])".format(alpha=IsAlpha), r"\1 ' \2"
    FR_IT_SPECIFIC_3 = u"([{alpha}])[']([^{alpha}])".format(alpha=IsAlpha), r"\1 ' \2"
    FR_IT_SPECIFIC_4 = u"([{alpha}])[']([{alpha}])".format(alpha=IsAlpha), r"\1' \2"

    FR_IT_SPECIFIC_APOSTROPHE = [FR_IT_SPECIFIC_1, FR_IT_SPECIFIC_2,
                                 FR_IT_SPECIFIC_3, FR_IT_SPECIFIC_4]

    NON_SPECIFIC_APOSTROPHE = r"\'", r" \' "

    MOSES_PENN_REGEXES_1 = [DEDUPLICATE_SPACE, ASCII_JUNK, DIRECTIONAL_QUOTE_1,
                            DIRECTIONAL_QUOTE_2, DIRECTIONAL_QUOTE_3,
                            DIRECTIONAL_QUOTE_4, DIRECTIONAL_QUOTE_5,
                            DIRECTIONAL_QUOTE_6, DIRECTIONAL_QUOTE_7,
                            DIRECTIONAL_QUOTE_8, REPLACE_ELLIPSIS, COMMA_1,
                            COMMA_2, COMMA_3, SYMBOLS, INTRATOKEN_SLASHES,
                            FINAL_PERIOD, PAD_QUESTION_EXCLAMATION_MARK,
                            PAD_PARENTHESIS, CONVERT_PARENTHESIS_1,
                            CONVERT_PARENTHESIS_2, CONVERT_PARENTHESIS_3,
                            CONVERT_PARENTHESIS_4, CONVERT_PARENTHESIS_5,
                            CONVERT_PARENTHESIS_6, PAD_DOUBLE_DASHES,
                            PAD_START_OF_STR, PAD_END_OF_STR,
                            CONVERT_DOUBLE_TO_SINGLE_QUOTES,
                            HANDLES_SINGLE_QUOTES, APOSTROPHE, CONTRACTION_1,
                            CONTRACTION_2, CONTRACTION_3, CONTRACTION_4,
                            CONTRACTION_5, CONTRACTION_6, CONTRACTION_7,
                            CONTRACTION_8, CONTRACTION_9, CONTRACTION_10,
                            CONTRACTION_11, CONTRACTION_12, CONTRACTION_13,
                            CONTRACTION_14, CONTRACTION_15, CONTRACTION_16,
                            CONTRACTION_17, CONTRACTION_18, CONTRACTION_19]

    MOSES_PENN_REGEXES_2 = [RESTORE_ELLIPSIS, CLEAN_EXTRA_SPACE_1,
                            CLEAN_EXTRA_SPACE_2, CLEAN_EXTRA_SPACE_3,
                            ESCAPE_AMPERSAND, ESCAPE_PIPE,
                            ESCAPE_LEFT_ANGLE_BRACKET, ESCAPE_RIGHT_ANGLE_BRACKET,
                            ESCAPE_SINGLE_QUOTE, ESCAPE_DOUBLE_QUOTE]

    MOSES_ESCAPE_XML_REGEXES = [ESCAPE_AMPERSAND, ESCAPE_PIPE,
                                ESCAPE_LEFT_ANGLE_BRACKET,
                                ESCAPE_RIGHT_ANGLE_BRACKET,
                                ESCAPE_SINGLE_QUOTE, ESCAPE_DOUBLE_QUOTE,
                                ESCAPE_LEFT_SQUARE_BRACKET,
                                ESCAPE_RIGHT_SQUARE_BRACKET]

    def __init__(self, lang='en'):
        # Initialize the object.
        super(MosesTokenizer, self).__init__()
        self.lang = lang
        # Initialize the language specific nonbreaking prefixes.
        self.NONBREAKING_PREFIXES = [_nbp.strip() for _nbp in nonbreaking_prefixes.words(lang)]
        self.NUMERIC_ONLY_PREFIXES = [w.rpartition(' ')[0] for w in
                                      self.NONBREAKING_PREFIXES if
                                      self.has_numeric_only(w)]

    def replace_multidots(self, text):
        text = re.sub(r'\.([\.]+)', r' DOTMULTI\1', text)
        while re.search(r'DOTMULTI\.', text):
            text = re.sub(r'DOTMULTI\.([^\.])', r'DOTDOTMULTI \1', text)
            text = re.sub(r'DOTMULTI\.', 'DOTDOTMULTI', text)
        return text

    def restore_multidots(self, text):
        while re.search(r'DOTDOTMULTI', text):
            text = re.sub(r'DOTDOTMULTI', r'DOTMULTI.', text)
        return re.sub(r'DOTMULTI', r'.', text)

    def islower(self, text):
        return not set(text).difference(set(self.IsLower))

    def isalpha(self, text):
        return not set(text).difference(set(self.IsAlpha))

    def has_numeric_only(self, text):
        return bool(re.search(r'(.*)[\s]+(\#NUMERIC_ONLY\#)', text))

    def handles_nonbreaking_prefixes(self, text):
        # Splits the text into tokens to check for nonbreaking prefixes.
        tokens = text.split()
        num_tokens = len(tokens)
        for i, token in enumerate(tokens):
            # Checks if token ends with a fullstop.
            token_ends_with_period = re.search(r'^(\S+)\.$', token)
            if token_ends_with_period:
                prefix = token_ends_with_period.group(1)
                # Checks for 3 conditions if
                # i.   the prefix contains a fullstop and
                #      any char in the prefix is within the IsAlpha charset
                # ii.  the prefix is in the list of nonbreaking prefixes and
                #      does not contain #NUMERIC_ONLY#
                # iii. the token is not the last token and that the
                #      next token contains all lowercase.
                if (('.' in prefix and self.isalpha(prefix)) or
                        (prefix in self.NONBREAKING_PREFIXES and
                         prefix not in self.NUMERIC_ONLY_PREFIXES) or
                        (i != num_tokens - 1 and self.islower(tokens[i + 1]))):
                    pass  # No change to the token.
                # Checks if the prefix is in NUMERIC_ONLY_PREFIXES
                # and ensures that the next word is a digit.
                elif (prefix in self.NUMERIC_ONLY_PREFIXES and
                      (i + 1) < num_tokens and
                      re.search(r'^[0-9]+', tokens[i + 1])):
                    pass  # No change to the token.
                else:  # Otherwise, adds a space after the tokens before a dot.
                    tokens[i] = prefix + ' .'
        return " ".join(tokens)  # Stitch the tokens back.

    def escape_xml(self, text):
        for regexp, substitution in self.MOSES_ESCAPE_XML_REGEXES:
            text = re.sub(regexp, substitution, text)
        return text

    def penn_tokenize(self, text, return_str=False):
        """
        This is a Python port of the Penn treebank tokenizer adapted by the Moses
        machine translation community. It's a little different from the
        version in nltk.tokenize.treebank.
        """
        # Converts input string into unicode.
        text = text_type(text)
        # Perform a chain of regex substituitions using MOSES_PENN_REGEXES_1
        for regexp, substitution in self.MOSES_PENN_REGEXES_1:
            text = re.sub(regexp, substitution, text)
        # Handles nonbreaking prefixes.
        text = self.handles_nonbreaking_prefixes(text)
        # Restore ellipsis, clean extra spaces, escape XML symbols.
        for regexp, substitution in self.MOSES_PENN_REGEXES_2:
            text = re.sub(regexp, substitution, text)
        return text if return_str else text.split()

    def tokenize(self, text, aggressive_dash_splits=False, return_str=False, escape=True):
        """
        Python port of the Moses tokenizer.
        >>> mtokenizer = MosesTokenizer()
        >>> text = u'Is 9.5 or 525,600 my favorite number?'
        >>> print (mtokenizer.tokenize(text, return_str=True))
        Is 9.5 or 525,600 my favorite number ?
        >>> text = u'The https://github.com/jonsafari/tok-tok/blob/master/tok-tok.pl is a website with/and/or slashes and sort of weird : things'
        >>> print (mtokenizer.tokenize(text, return_str=True))
        The https : / / github.com / jonsafari / tok-tok / blob / master / tok-tok.pl is a website with / and / or slashes and sort of weird : things
        >>> text = u'This, is a sentence with weird\xbb symbols\u2026 appearing everywhere\xbf'
        >>> expected = u'This , is a sentence with weird \xbb symbols \u2026 appearing everywhere \xbf'
        >>> assert mtokenizer.tokenize(text, return_str=True) == expected
        :param tokens: A single string, i.e. sentence text.
        :type tokens: str
        :param aggressive_dash_splits: Option to trigger dash split rules .
        :type aggressive_dash_splits: bool
        """
        # Converts input string into unicode.
        text = text_type(text)

        # De-duplicate spaces and clean ASCII junk
        for regexp, substitution in [self.DEDUPLICATE_SPACE, self.ASCII_JUNK]:
            text = re.sub(regexp, substitution, text)
        # Strips heading and trailing spaces.
        text = text.strip()
        # Separate special characters outside of IsAlnum character set.
        regexp, substitution = self.PAD_NOT_ISALNUM
        text = re.sub(regexp, substitution, text)
        # Aggressively splits dashes
        if aggressive_dash_splits:
            regexp, substitution = self.AGGRESSIVE_HYPHEN_SPLIT
            text = re.sub(regexp, substitution, text)
        # Replaces multidots with "DOTDOTMULTI" literal strings.
        text = self.replace_multidots(text)
        # Separate out "," except if within numbers e.g. 5,300
        for regexp, substitution in [self.COMMA_SEPARATE_1, self.COMMA_SEPARATE_2]:
            text = re.sub(regexp, substitution, text)

        # (Language-specific) apostrophe tokenization.
        if self.lang == 'en':
            for regexp, substitution in self.ENGLISH_SPECIFIC_APOSTROPHE:
                text = re.sub(regexp, substitution, text)
        elif self.lang in ['fr', 'it']:
            for regexp, substitution in self.FR_IT_SPECIFIC_APOSTROPHE:
                text = re.sub(regexp, substitution, text)
        else:
            regexp, substitution = self.NON_SPECIFIC_APOSTROPHE
            text = re.sub(regexp, substitution, text)

        # Handles nonbreaking prefixes.
        text = self.handles_nonbreaking_prefixes(text)
        # Cleans up extraneous spaces.
        regexp, substitution = self.DEDUPLICATE_SPACE
        text = re.sub(regexp, substitution, text).strip()
        # Restore multidots.
        text = self.restore_multidots(text)
        if escape:
            # Escape XML symbols.
            text = self.escape_xml(text)

        return text if return_str else text.split()


'''
Functions that tokenise text and returns the list of tokens all of which are
Strings.

Functions:

1. :py:func:`bella.tokenisers.whitespace` -- tokenises on whitespace.
2. :py:func:`bella.tokenisers.ark_twokenize` -- A Twitter tokeniser from
   `CMU Ark <https://github.com/brendano/ark-tweet-nlp>`_
3. :py:func:`bella.tokenisers.stanford` -- Stanford tokeniser from
   `CoreNLP <https://nlp.stanford.edu/software/tokenizer.html>`_
4. :py:func:`bella.tokenisers.moses` -- Tokeniser used in the
   `moses toolkit <https://github.com/moses-smt>`_
'''
from typing import List

import twokenize

from bella import stanford_tools
from bella.moses_tools import MosesTokenizer


def whitespace(text: str) -> List[str]:
    '''
    Tokenises on whitespace.

    :param text: A string to be tokenised.
    :returns: A list of tokens where each token is a String.
    '''

    if isinstance(text, str):
        return text.split()
    raise ValueError(f'The paramter must be of type str not {type(text)}')


def ark_twokenize(text: str) -> List[str]:
    '''
    A Twitter tokeniser from
    `CMU Ark <https://github.com/brendano/ark-tweet-nlp>`_

    This is a wrapper of
    `this <https://github.com/Sentimentron/ark-twokenize-py>`_

    :param text: A string to be tokenised.
    :returns: A list of tokens where each token is a String.
    '''

    if isinstance(text, str):
        return twokenize.tokenizeRawTweetText(text)
    raise ValueError(f'The paramter must be of type str not {type(text)}')


def stanford(text: str) -> List[str]:
    '''
    Stanford tokeniser from
    `CoreNLP <https://nlp.stanford.edu/software/tokenizer.html>`_

    Requires CoreNLP server to be running.

    :param text: A string to be tokenised.
    :returns: A list of tokens where each token is a String.
    '''
    if isinstance(text, str):
        return stanford_tools.tokenise(text)
    raise ValueError(f'The paramter must be of type str not {type(text)}')


def moses(text: str, aggressive_dash_splits: bool = False, 
          escape: bool = True) -> List[str]:
    '''
    Tokeniser used in the `moses toolkit <https://github.com/moses-smt>`_

    :param text: A string to be tokenised.
    :param aggressive_dash_splits: Option to trigger dash split rules
    :param escape: Whether to escape characters e.g. "'s" escaped equals 
                   "&apos;s"
    :returns: A list of tokens where each token is a String.
    '''

    if isinstance(text, str):
        moses = MosesTokenizer()
        return moses.tokenize(text, 
                              aggressive_dash_splits=aggressive_dash_splits,
                              escape=escape)
    raise ValueError(f'The paramter must be of type str not {type(text)}')
